@@ -310,7 +310,6 @@ def get_esm_embedd(seq, embedd_model, batch_converter, msa_data=None):
     token_reps = results["representations"][REPR_LAYER_NUM][..., 1:, :].unsqueeze(dim=1)
     return token_reps
 
-
 def get_all_protein_ids(dataloader, verbose=False):
     """ Given a sidechainnet dataloader for a CASP version, 
         Returns all the ids belonging to proteins.
@@ -492,7 +491,7 @@ def prot_covalent_bond(seqs, adj_degree=1, cloud_mask=None, mat=True):
         # convert to undirected
         adj_mat[s] = adj_mat[s] + adj_mat[s].t()
         # do N_th degree adjacency
-        adj_mat, attr_mat = nth_deg_adjacency(adj_mat, n=adj_degree, sparse=True)
+        adj_mat, attr_mat = nth_deg_adjacency(adj_mat, n=adj_degree, sparse=False) # True
 
     if mat: 
         return attr_mat.bool().to(seqs.device), attr_mat.to(device)
@@ -642,36 +641,36 @@ def mds_torch(pre_dist_mat, weights=None, iters=10, tol=1e-5, eigen=False, verbo
         * historic_stresses: (batch x steps)
     """
     device, dtype = pre_dist_mat.device, pre_dist_mat.type()
-
+    # ensure batched MDS
+    pre_dist_mat = expand_dims_to(pre_dist_mat, length = ( 3 - len(pre_dist_mat.shape) ))
     # start
     batch, N, _ = pre_dist_mat.shape
     diag_idxs = np.arange(N)
     his = [torch.tensor([np.inf]*batch, device=device)]
 
-    # do it by eigendecomposition - way faster but not weights
-    # https://www.biorxiv.org/content/10.1101/2020.11.27.401232v1.full.pdf
-    if eigen == True and weights is None:
-        preds_3d = []
-        for bi in range(pre_dist_mat.shape[0]):
-            D = pre_dist_mat[bi]**2
-            M = D[:1, :] + D[:, :1] - D 
-            u,s,v = torch.svd_lowrank(M/2)
-            preds_3d.append( (u@torch.diag(s).sqrt())[:, :3].t() )
-        return torch.stack(preds_3d, dim=0), torch.zeros_like(torch.stack(his, dim=0))
-    elif eigen == True:
-        if verbose: 
+    # initialize by eigendecomposition: https://www.lptmc.jussieu.fr/user/lesne/bioinformatics.pdf
+    # follow : https://www.biorxiv.org/content/10.1101/2020.11.27.401232v1.full.pdf
+    D = pre_dist_mat**2
+    M =  0.5 * (D[:, :1, :] + D[:, :, :1] - D) 
+    # do loop svd bc it's faster: (2-3x in CPU and 1-2x in GPU)
+    # https://discuss.pytorch.org/t/batched-svd-lowrank-being-much-slower-than-loop-implementation-both-cpu-and-gpu/119336
+    svds = [torch.svd_lowrank(mi) for mi in M]
+    u = torch.stack([svd[0] for svd in svds], dim=0)
+    s = torch.stack([svd[1] for svd in svds], dim=0)
+    v = torch.stack([svd[2] for svd in svds], dim=0)
+    best_3d_coords = torch.bmm(u, torch.diag_embed(s).sqrt())[..., :3]
+
+    # only eigen - way faster but not weights
+    if weights is None and eigen==True:
+        return torch.transpose( best_3d_coords, -1, -2), torch.zeros_like(torch.stack(his, dim=0))
+    elif eigen==True:
+        if verbose:
             print("Can't use eigen flag if weights are active. Fallback to iterative")
 
     # continue the iterative way
     if weights is None:
         weights = torch.ones_like(pre_dist_mat)
 
-    # ensure batched MDS
-    pre_dist_mat = expand_dims_to(pre_dist_mat, length = ( 3 - len(pre_dist_mat.shape) ))
-
-    # init random coords
-    best_stress = float("Inf") * torch.ones(batch, device = device).type(dtype)
-    best_3d_coords = 2*torch.rand(batch, N, 3, device = device).type(dtype) - 1
     # iterative updates:
     for i in range(iters):
         # compute distance matrix of coords and stress
@@ -686,16 +685,16 @@ def mds_torch(pre_dist_mat, weights=None, iters=10, tol=1e-5, eigen=False, verbo
         # update
         coords = (1. / N * torch.matmul(B, best_3d_coords))
         dis = torch.norm(coords, dim=(-1, -2))
+
         if verbose >= 2:
             print('it: %d, stress %s' % (i, stress))
         # update metrics if relative improvement above tolerance
-        if (best_stress - stress / dis).mean() <= tol:
+        if (his[-1] - stress / dis).mean() <= tol:
             if verbose:
                 print('breaking at iteration %d with stress %s' % (i,
                                                                    stress / dis))
             break
 
-        pre_dist_mat = dist_mat
         best_3d_coords = coords
         his.append( stress / dis )
 
@@ -877,16 +876,21 @@ def kabsch_torch(X, Y, cpu=True):
     C = torch.matmul(X_, Y_.t()).detach()
     if cpu: 
         C = C.cpu()
-    # Optimal rotation matrix via SVD - warning! W must be transposed
-    svd_torch = torch.svd if int(torch.__version__.split(".")[1]) < 8 else torch.linalg.svd
-    V, S, W = svd_torch(C)
+    # Optimal rotation matrix via SVD
+    if int(torch.__version__.split(".")[1]) < 8:
+        # warning! int torch 1.<8 : W must be transposed
+        V, S, W = torch.svd(C)
+        W = W.t()
+    else: 
+        V, S, W = torch.linalg.svd(C)
+    
     # determinant sign for direction correction
     d = (torch.det(V) * torch.det(W)) < 0.0
     if d:
         S[-1]    = S[-1] * (-1)
         V[:, -1] = V[:, -1] * (-1)
     # Create Rotation matrix U
-    U = torch.matmul(V, W.t()).to(device)
+    U = torch.matmul(V, W).to(device)
     # calculate rotations
     X_ = torch.matmul(X_.t(), U).t()
     # return centered and aligned
@@ -1010,17 +1014,17 @@ def tmscore_numpy(X, Y):
 
 def mdscaling_torch(pre_dist_mat, weights=None, iters=10, tol=1e-5,
                     fix_mirror=True, N_mask=None, CA_mask=None, C_mask=None, 
-                    eigen=True, verbose=2):
+                    eigen=False, verbose=2):
     """ Handles the specifics of MDS for proteins (mirrors, ...) """
     # batched mds for full parallel 
     preds, stresses = mds_torch(pre_dist_mat, weights=weights,iters=iters, 
-                                              tol=tol, eigen=True, verbose=verbose)
+                                              tol=tol, eigen=eigen, verbose=verbose)
     if not fix_mirror:
         return preds, stresses
 
     # no need to caculate multiple mirrors - just correct Z axis
     phi_ratios = calc_phis_torch(preds, N_mask, CA_mask, C_mask, prop=True)
-    to_correct = torch.nonzero( (phi_ratios < 0.5) ).view(-1)
+    to_correct = torch.nonzero( (phi_ratios < 0.5)).view(-1)
     # fix mirrors by (-1)*Z if more (+) than (-) phi angles
     preds[to_correct, -1] = (-1)*preds[to_correct, -1]
     if verbose == 2:
